(* Barrier theorem: how to restore a strong model on a weaker architecture *)
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice.
Require Import  path fintype finset fingraph.
Require Import extrarel wmm.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section FixedSize.

Variable n : nat.

Notation "'Word" := (Word n).
Notation "'Address" := (Address n).
Notation "'Value" := (Value n).
Notation "'Proc" := (Proc n).
Notation "'Eiid" := (Eiid n).
Notation "'Location" := (Location n).
Notation "'Event" := (Event n).
Notation "'Event_struct" := (Event_struct n).
Notation "'Action" := (Action n).
Notation "'Execution_witness" := (Execution_witness n).
Notation "'Archi":= (Archi n).

Section ArchiOrdering.

Definition weaker (A1 A2 : 'Archi) := 
  (forall E, subrel (ppo A1 E) (ppo A2 E)) /\
  forall X,  subrel (grf A1 X) (grf A2 X).

Definition naked_abs (E : 'Event_struct)(X : 'Execution_witness) : rel 'Event :=
  fun x y => false.

Lemma naked_evts : forall E X, 
  subrel (naked_abs E X) [rel x y \in (events_of_ES E)].
Proof. by move=> E X e1 e2; rewrite /naked_abs. Qed.

Definition nakedArchi (A : 'Archi) :=
  mkArchi (@grf_valid _ A)
  (@ppo_valid _ A) naked_evts.

Definition rf_sub (A1 A2 : 'Archi)(X : 'Execution_witness) : rel 'Event :=
  [rel e1 e2 | (grf A2 X e1 e2) && ~~(grf A1 X e1 e2)].

Definition ab (A1 A2 : 'Archi)(E : 'Event_struct) (X : 'Execution_witness)
  (fenb fenc : rel 'Event) : rel 'Event :=
  [rel e1 e2 | 
    (fenb e1 e2) || existsb ew, (rf_sub A1 A2 X) e1 ew && (fenc ew e2)].


Definition ppo_sub (A1 A2 : 'Archi)(E : 'Event_struct) : rel 'Event :=
  [rel e1 e2 | (ppo A2 E e1 e2) && ~~(ppo A1 E e1 e2)].

End ArchiOrdering.

Section Barriers.

Lemma acyclicity_crit (F : finType) (r1 r2 : rel F) (ac_r1 : acyclic r1)
  (crit : forall x y z, r2 x y -> r2 y z -> r1 x y || r1 x z) :
  acyclic r2.
Proof.
move=> l2 sl2 _ cl2; case: {-1}l2 (erefl l2) sl2 cl2 => // [a l hl2 _] cl2.
have ha: a \in l2 by rewrite hl2 in_cons eqxx.
rewrite -hl2 in cl2 => {hl2}.
case: (@pigeon_cycle _ r1 (mem l2) a)=> //.
  move=> x l2x; pose x' := next l2 x; pose x'' := next l2 x'.
  case: (orP (crit x x' x'' _ _)). 
  * by apply: next_cycle.
  * by apply: next_cycle; rewrite // mem_next.
  * by move=> hx'; exists x'; rewrite // mem_next.
  by move=> hx''; exists x''; rewrite // !mem_next.
by move=> t /andP [st l2t]; apply: ac_r1=> //; apply/subsetP.
Qed.

Lemma barriers_placement :
  forall (A1 A2 : 'Archi)(E : 'Event_struct)(X : 'Execution_witness), 
  dp_of_ES E =2 dp_of_ES E ->
  weaker A1 A2 -> wf_ES E ->
  abc A1 E X =2 ab A1 A2 E X (ppo_sub A1 A2 E) (ppo A2 E) ->
  wf_EW E X ->
  (valid_execution A1 E X ->
  valid_execution (nakedArchi A2) E X).
Proof.
move=>A1 A2 E X hdp [weak_ppo weak_grf].
move=> [hsameid [hdp_poiico hdp_reads] [hiico_id hiico_e  hiico_c]].
move=> hab [wf_ws wf_rf] [hb_pio_c rf_dp_c ghb_c]; split=> //.
case: wf_rf=> _ hrinj _ _ hrfmaps; case: wf_ws=> [ _ [_ _ hwsmaps]].
set r1 : rel 'Event := (ghb A1 E X).
set r2 : rel 'Event := (ghb (nakedArchi A2) E X).
apply: (acyclicity_crit ghb_c).
move=> x y z => r2xy r2yz; case r1xy: (r1 x y); first by rewrite -/r1 r1xy.
move/negbT : r1xy =>  nr1xy.
case/orP: r2xy => /=; last first.
  move: nr1xy; rewrite !negb_or; case/and4P=> nr11 nr12; case/andP; move/negPf->.
  move/negPf=> -> /= nr13 ppo2xy; move: nr12; rewrite hab !negb_or /ppo_sub /=.
  by rewrite ppo2xy /= negbK; move/negPf: nr13->.
move=> grf2xy.
have rf_subxy : rf_sub A1 A2 X x y. 
  by rewrite /rf_sub /= grf2xy /=; move: nr1xy; rewrite !negb_or; case/andP.
move/grf_valid: (grf2xy); move/hrfmaps; move/rfmaps_isRr => diry.
case/orP: r2yz => /=.
   by move/grf_valid; move/hrfmaps; move/rfmaps_isWl; rewrite diry.
case c1 : (ws_of_EW X y z) => /=.
  by move/hwsmaps: c1; move/wsmaps_isWl; rewrite diry.
case c2: (fr X y z) => /=.
  move=> _; rewrite /ghb /= hab /ab /=.
  case/existsP: c2=> t; case/andP; case=> hl hr.
  move/grf_valid: grf2xy => hl'; have := (hrinj _ _ _ hl hl') => etx.
  by rewrite -etx hr !orbT.
move=> ppo2yz; rewrite /ghb /=.
suff -> : abc A1 E X x z by rewrite !orbT.
rewrite hab /ab /=.
suff -> : existsb ew, rf_sub A1 A2 X x ew && ppo A2 E ew z by rewrite orbT.
apply/existsP; exists y; rewrite /rf_sub /= grf2xy /=; move: nr1xy.
by rewrite !negb_or; case/andP=> ->.
Qed.



End Barriers.

End FixedSize.


